/*
 * Copyright (c) 2017 Ted Unangst <tedu@tedunangst.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
package main

import (
	"bufio"
	"bytes"
	"compress/gzip"
	"encoding/json"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"text/template"

	// brotlidec "gopkg.in/kothar/brotli-go.v0/dec"
	"github.com/andybalholm/cascadia"
	"golang.org/x/net/html"
)

var permittedtags = []string{"title", "div", "span", "h1", "h2", "h3", "h4", "h5", "h6",
	"table", "thead", "tbody", "th", "tr", "td",
	"p", "br", "pre", "code", "blockquote", "strong", "em", "b", "i", "s", "sup",
	"ol", "ul", "li"}
var permittedattr = []string{"colspan", "rowspan"}
var bannedtags = []string{"script", "style"}

func init() {
	sort.Strings(permittedtags)
	sort.Strings(permittedattr)
	sort.Strings(bannedtags)
}

func contains(array []string, tag string) bool {
	idx := sort.SearchStrings(array, tag)
	return idx < len(array) && array[idx] == tag
}

func getattr(node *html.Node, attr string) string {
	for _, a := range node.Attr {
		if a.Key == attr {
			return a.Val
		}
	}
	return ""
}

func writetag(w io.Writer, node *html.Node) {
	io.WriteString(w, "<")
	io.WriteString(w, node.Data)
	for _, attr := range node.Attr {
		if contains(permittedattr, attr.Key) {
			io.WriteString(w, fmt.Sprintf(` %s="%s"`, attr.Key,
				html.EscapeString(attr.Val)))
		}
	}
	io.WriteString(w, ">")
}

func jsonfind(jmap map[string]interface{}, keys []string) string {
	var ii interface{}
	ii = jmap
	for _, key := range keys {
		idx, err := strconv.Atoi(key)
		if err == nil {
			m := ii.([]interface{})
			ii = m[idx]
		} else {
			m := ii.(map[string]interface{})
			ii = m[key]
		}
	}
	return ii.(string)
}

func extractblogshit(script string, baseurl *url.URL) *html.Node {
	re := regexp.MustCompile("blogspot.*'postId': '([0-9]+)'")
	m := re.FindStringSubmatch(script)
	if m == nil {
		return nil
	}
	targ := fmt.Sprintf("%s://%s/feeds/posts/default/%s?alt=json", baseurl.Scheme, baseurl.Hostname(), m[1])
	log.Printf("now i'm getting real content from %s", targ)
	jr, err := http.Get(targ)
	if err != nil {
		log.Println("json post err", err)
		return nil
	}
	defer jr.Body.Close()
	decoder := json.NewDecoder(jr.Body)
	var jmap map[string]interface{}
	err = decoder.Decode(&jmap)
	if err != nil {
		log.Println("json decode error")
		return nil
	}
	body := jsonfind(jmap, []string{"entry", "content", "$t"})
	root, _ := html.Parse(strings.NewReader(body))
	return root
}

func procscriptnode(w io.Writer, script *html.Node, baseurl *url.URL) {
	src := getattr(script, "src")
	if src != "" {
		scripturl, err := url.Parse(src)
		if err != nil {
			log.Println("frame err", err)
			return
		}
		if scripturl.Host == "gist.github.com" {
			path := scripturl.Path
			if strings.HasSuffix(path, ".js") {
				path = path[:len(path)-3] + "/raw"
			}
			scripturl.Path = path
			txt, err := http.Get(scripturl.String())
			if err != nil {
				log.Println("frame err", err)
				return
			}
			defer txt.Body.Close()
			var buf bytes.Buffer
			io.Copy(&buf, txt.Body)
			io.WriteString(w, `<pre style="border: 2px solid black; padding: 1em;">`)
			template.HTMLEscape(w, buf.Bytes())
			io.WriteString(w, `</pre>`)
		}
	} else {
		if c := script.FirstChild; c != nil && c.Type == html.TextNode {
			article := extractblogshit(c.Data, baseurl)
			if article != nil {
				clean(w, article, baseurl)
			}
		}
	}
}

func insertframecontents(w io.Writer, node *html.Node, baseurl *url.URL) {
	src := getattr(node, "src")
	if src == "" {
		return
	}
	u, err := url.Parse(src)
	if err != nil {
		log.Println("frame err", err)
		return
	}
	targ := baseurl.ResolveReference(u)
	if targ.Hostname() == "www.youtube.com" {
		hint := "A youtube video"
		th := getattr(node, "data-thumbnail-src")
		if th != "" {
			hint = fmt.Sprintf(`<img src="%s">`, html.EscapeString(th))
		}
		io.WriteString(w, fmt.Sprintf(`<a href="%s">%s</a>`, html.EscapeString(targ.String()), hint))
		return
	}
	fr, err := http.Get(targ.String())
	if err != nil {
		log.Println("frame err", err)
		return
	}
	defer fr.Body.Close()
	if strings.HasPrefix(fr.Header.Get("Content-Type"), "text/html") {
		root, err := html.Parse(fr.Body)
		if err != nil {
			log.Println("frame err", err)
			return
		}
		sel := cascadia.MustCompile("script")
		scripts := sel.MatchAll(root)
		for _, script := range scripts {
			procscriptnode(w, script, baseurl)
		}
		blocksel := cascadia.MustCompile("blockquote")
		quotes := blocksel.MatchAll(root)
		for _, quote := range quotes {
			clean(w, quote, targ)
		}
	}
}

func haschild(node *html.Node, css string) bool {
	sel := cascadia.MustCompile(css)
	return sel.MatchFirst(node) != nil
}

func clean(w io.Writer, node *html.Node, baseurl *url.URL) {
	switch node.Type {
	case html.ElementNode:
		tag := node.Data
		switch {
		case tag == "a":
			io.WriteString(w, fmt.Sprintf(`<a href="%s">`,
				html.EscapeString(getattr(node, "href"))))
		case tag == "img":
			src := getattr(node, "src")
			if getattr(node, "data-native-src") != "" {
				src = getattr(node, "data-native-src")
				if strings.HasSuffix(src, "-1x-1.jpg") {
					src = src[:len(src)-9] + "1200x-1.jpg"
				}
			} else if getattr(node, "data-raw-src") != "" {
				src = getattr(node, "data-raw-src")
			}

			if strings.Contains(getattr(node, "class"), "progressiveMedia-thumbnail") {
				src = ""
			}
			if src != "" {
				io.WriteString(w, fmt.Sprintf(`<img src="%s">`,
					html.EscapeString(src)))
			}
		case tag == "picture":
			src := ""
			for c := node.FirstChild; c != nil; c = c.NextSibling {
				if c.Type == html.ElementNode && c.Data == "source" {
					attr := getattr(c, "data-srcset")
					if attr != "" {
						src = attr
					}
				}
			}
			if src != "" {
				io.WriteString(w, fmt.Sprintf(`<picture><img src="%s"></picture>`,
					html.EscapeString(src)))
				return
			}
		case tag == "figure":
			src := getattr(node, "itemid")
			if src != "" {
				if !haschild(node, "img") {
					io.WriteString(w, fmt.Sprintf(`<img src="%s">`,
						html.EscapeString(src)))
				}
			}
		case tag == "iframe":
			insertframecontents(w, node, baseurl)
		case tag == "noscript":
			// see https://github.com/golang/go/issues/16318
			if c := node.FirstChild; c != nil {
				rereader := strings.NewReader(c.Data)
				root, err := html.Parse(rereader)
				if err == nil {
					clean(w, root, baseurl)
				}
			}
			return
		case tag == "script":
			procscriptnode(w, node, baseurl)
			return
		case tag == "link":
			rel := getattr(node, "rel")
			if rel == "alternate" || rel == "icon" {
				html.Render(w, node)
			}
		case contains(permittedtags, tag):
			writetag(w, node)
		case contains(bannedtags, tag):
			return
		}
	case html.TextNode:
		io.WriteString(w, html.EscapeString(node.Data))
	}
	for c := node.FirstChild; c != nil; c = c.NextSibling {
		clean(w, c, baseurl)
	}
	if node.Type == html.ElementNode {
		tag := node.Data
		if tag == "a" || (contains(permittedtags, tag) && tag != "br") {
			io.WriteString(w, fmt.Sprintf("</%s>", tag))
		}
		if tag == "p" || tag == "div" {
			io.WriteString(w, "\n")
		}
	}
}

func filter(L *Interpreter, w io.Writer, resp *http.Response, req *http.Request) {
	if resp.StatusCode != 200 {
		resp.Write(w)
		return
	}
	if !strings.HasPrefix(resp.Header.Get("Content-Type"), "text/html") {
		resp.Write(w)
		return
	}
	log.Printf("considering filtering %s %s", req.Host, req.URL.Path)

	reader := resp.Body
	defer reader.Close()

	gzipit := false
	if strings.EqualFold(resp.Header.Get("Content-Encoding"), "gzip") {
		gzr, err := gzip.NewReader(reader)
		if err != nil {
			log.Printf("error unzipping: %s", err)
			// now what?
			return
		}
		reader = gzr
		defer reader.Close()
		gzipit = true
	}

	respreader, respwriter := io.Pipe()
	var gzw flushWriter
	if gzipit {
		gzw = gzip.NewWriter(respwriter)
	} else {
		gzw = bufio.NewWriter(respwriter)
	}

	go dofiltering(L, respwriter, gzw, reader, req)

	resp.Header.Set("Content-Type", "text/html; charset=utf-8")
	resp.Header.Del("Content-Length")
	resp.ContentLength = -1
	resp.TransferEncoding = []string{"chunked"}
	resp.Body = respreader
	// respreader will block until dofiltering closes the respwriter
	// don't need to worry about returning before filtering is done
	resp.Write(w)
}

type flushWriter interface {
	Write(p []byte) (int, error)
	Flush() error
}

func dofiltering(L *Interpreter, under io.WriteCloser, w flushWriter, reader io.Reader,
	req *http.Request) {
	defer under.Close()
	wcloser, ok := w.(io.Closer)
	if ok {
		defer wcloser.Close()
	}
	defer w.Flush()

	root, err := html.Parse(reader)
	if err != nil {
		log.Printf("error parsing html from %s: %s", req.URL.Path, err)
		io.WriteString(w, "HTML could not be parsed")
		return
	}

	filtresult := L.Call("filter", w, req, req.URL.Hostname(), req.URL.Path, root)
	if filtresult == "" {
		log.Printf("did not rewrite %s, raw rendering", req.URL.String())
		html.Render(w, root)
		return
	}
}

func prefilter(L *Interpreter, req *http.Request) {
	L.Call("prefilter", req)
}

func shouldintercept(L *Interpreter, host string) bool {
	filtresult := L.Call("shouldfilter", host)
	if filtresult != "" {
		return true
	}
	return false
}
