/*
 * Copyright (c) 2017 Ted Unangst <tedu@tedunangst.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
package main

import (
	"bufio"
	"context"
	"crypto/rand"
	"crypto/rsa"
	"crypto/tls"
	"crypto/x509"
	"crypto/x509/pkix"
	"io"
	"log"
	"math/big"
	"net"
	"net/http"
	"sync"
	"time"

	"golang.org/x/net/proxy"
)

func copyuntildone(w net.Conn, r net.Conn, pipe chan net.Conn) {
	io.Copy(w, r)
	pipe <- r
}

func closewhendone(pipe chan net.Conn) {
	w := <-pipe
	r := <-pipe
	w.Close()
	r.Close()
	close(pipe)
}

func noRedirects(req *http.Request, via []*http.Request) error {
	return http.ErrUseLastResponse
}

type contextDialer interface {
	DialContext(ctx context.Context, network, address string) (net.Conn, error)
}

// get a dialer to connect out
func getDialer() contextDialer {
	usetor := false

	if usetor {
		// the internal socks dialer implements context functions,
		// but the exported proxy.Dialer type does not
		dialer, _ := proxy.SOCKS5("tcp", "127.0.0.1:9050", nil, nil)
		return dialer.(contextDialer)
	}
	return &net.Dialer{}
}

// this should not be different from the TLS intercept proxy
func proxyreq(w http.ResponseWriter, r *http.Request) {
	deadline := time.Now().Add(1 * time.Minute)
	ctx, cancel := context.WithDeadline(r.Context(), deadline)
	defer cancel()
	log.Printf("got a request %s\n", r.URL.String())
	if r.URL.Port() == "" {
		r.URL.Host = r.URL.Host + ":" + "80"
	}
	dialer := getDialer()
	serverconn, err := dialer.DialContext(ctx, "tcp", r.URL.Host)
	if err != nil {
		log.Printf("err proxying %s", err)
		http.Error(w, "no proxying", http.StatusInternalServerError)
		return
	}
	r.Write(serverconn)

	hj, ok := w.(http.Hijacker)
	if !ok {
		http.Error(w, "no hijacking", http.StatusInternalServerError)
		return
	}
	clientconn, _, _ := hj.Hijack()
	// clientconn.SetDeadline(deadline)
	defer clientconn.Close()

	outbuf := bufio.NewReader(serverconn)
	resp, err := http.ReadResponse(outbuf, r)
	if err != nil {
		log.Printf("failed to read response %s", err)
		writeerror(clientconn, "failed to read response")
		return
	}

	L := getinterpreter()
	defer putinterpreter(L)
	if !shouldintercept(L, r.URL.Hostname()) {
		resp.Write(clientconn)
		return
	}
	filter(L, clientconn, resp, r)
}

type expiringConfig struct {
	config *tls.Config
	expiry time.Time
}

var certcache = make(map[string]*expiringConfig)
var certcachemtx sync.Mutex

func certexpiry() {
	c := time.Tick(5 * time.Minute)
	for now := range c {
		deadline := now.Add(10 * time.Minute)
		certcachemtx.Lock()
		for h, conf := range certcache {
			if conf.expiry.Before(deadline) {
				log.Printf("cert for %s is too old", h)
				delete(certcache, h)
			}
		}
		certcachemtx.Unlock()
	}
}

func getTLSConfig(host string) *tls.Config {
	certcachemtx.Lock()
	expconf := certcache[host]
	certcachemtx.Unlock()
	if expconf != nil {
		return expconf.config
	}

	cacert, err := tls.LoadX509KeyPair("intercept.crt", "intercept.key")
	if err != nil {
		log.Fatalf("can't load intercept ca")
	}

	ca509, err := x509.ParseCertificate(cacert.Certificate[0])
	if err != nil {
		log.Fatalf("Failed to parse certificate: %s", err)
	}

	rsaBits := 1024
	priv, err := rsa.GenerateKey(rand.Reader, rsaBits)
	if err != nil {
		log.Fatalf("Failed to create key: %s", err)
	}

	notBefore := time.Now()
	notAfter := notBefore.Add(time.Hour)

	serialNumberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, serialNumberLimit)
	if err != nil {
		log.Fatalf("failed to generate serial number: %s", err)
	}

	template := x509.Certificate{
		SerialNumber: serialNumber,
		Issuer:       ca509.Subject,
		Subject: pkix.Name{
			Organization: []string{"intercepted"},
		},
		NotBefore: notBefore,
		NotAfter:  notAfter,

		KeyUsage:              x509.KeyUsageKeyEncipherment | x509.KeyUsageDigitalSignature,
		ExtKeyUsage:           []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth},
		BasicConstraintsValid: true,
	}

	template.DNSNames = append(template.DNSNames, host)

	derBytes, err := x509.CreateCertificate(rand.Reader, &template, ca509, &priv.PublicKey,
		cacert.PrivateKey)
	if err != nil {
		log.Fatalf("Failed to create certificate: %s", err)
	}

	expconf = &expiringConfig{
		config: &tls.Config{
			Certificates: []tls.Certificate{
				tls.Certificate{Certificate: [][]byte{derBytes, cacert.Certificate[0]},
					PrivateKey: priv,
				}},
		},
		expiry: notAfter,
	}
	certcachemtx.Lock()
	certcache[host] = expconf
	certcachemtx.Unlock()
	return expconf.config
}

type Proxer struct {
}

func writeerror(w io.Writer, s string) {
	io.WriteString(w, "HTTP/1.0 503 Error\r\n\r\n")
	io.WriteString(w, s)
}

func (pxr *Proxer) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	deadline := time.Now().Add(1 * time.Minute)
	ctx, cancel := context.WithDeadline(r.Context(), deadline)
	defer cancel()
	if r.Method != "CONNECT" {
		proxyreq(w, r)
		return
	}
	dest := r.URL.Host
	desthost := r.URL.Hostname()
	log.Printf("got a connect to %s", dest)
	hj, ok := w.(http.Hijacker)
	if !ok {
		http.Error(w, "no hijacking", http.StatusInternalServerError)
		return
	}
	clientconn, _, err := hj.Hijack()
	if err != nil {
		log.Printf("error hijacking %s", err)
		return
	}

	log.Printf("dialing %s\n", dest)
	serverconn, err := getDialer().DialContext(ctx, "tcp", dest)
	if err != nil {
		log.Printf("failed to connect %s", err)
		writeerror(clientconn, "failed to connect")
		return
	}

	L := getinterpreter()
	defer putinterpreter(L)
	if !shouldintercept(L, desthost) {
		clientconn.Write([]byte("HTTP/1.0 200 OK\r\n\r\n"))
		p := make(chan net.Conn)
		go copyuntildone(clientconn, serverconn, p)
		go copyuntildone(serverconn, clientconn, p)
		go closewhendone(p)
		return
	}
	serverconn = tls.Client(serverconn, &tls.Config{ServerName: desthost})
	defer serverconn.Close()
	err = serverconn.(*tls.Conn).Handshake()
	if err != nil {
		log.Printf("handshake error: %s", err)
		writeerror(clientconn, "unable to make secure connection")
		return
	}
	// serverconn.SetDeadline(deadline)

	clientconn.Write([]byte("HTTP/1.0 200 OK\r\n\r\n"))
	tlsconf := getTLSConfig(desthost)
	clientconn = tls.Server(clientconn, tlsconf)
	// clientconn.SetDeadline(deadline)
	defer clientconn.Close()

	clientreq, err := http.ReadRequest(bufio.NewReader(clientconn))
	if err != nil {
		log.Printf("failed to read request %s", err)
		writeerror(clientconn, "failed to read request")
		return
	}
	clientreq.Host = ""
	clientreq.URL.Scheme = "https"
	clientreq.URL.Host = dest
	clientreq.Close = true
	clientreq.Header.Set("Connection", "close")
	if clientreq.Header.Get("Accept-Encoding") != "" {
		clientreq.Header.Set("Accept-Encoding", "gzip")
	}
	log.Printf("clientreq url %s\n", clientreq.URL.String())
	prefilter(L, clientreq)
	clientreq.URL.Host = clientreq.URL.Hostname()

	log.Printf("connection success")
	clientreq.Write(serverconn)
	serverresp, err := http.ReadResponse(bufio.NewReader(serverconn), clientreq)
	if err != nil {
		log.Printf("failed to read response %s", err)
		writeerror(clientconn, "failed to read response")
		return
	}
	serverresp.Close = true
	serverresp.Header.Set("Connection", "close")
	filter(L, clientconn, serverresp, clientreq)
}

func main() {
	go certexpiry()
	var pxr Proxer
	http.ListenAndServe("127.0.0.1:8090", &pxr)
}
